"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
exports.__esModule = true;
exports.importAliasRule = void 0;
var alias_config_1 = require("#src/alias-config");
var fs_extra_1 = require("fs-extra");
var path_1 = require("path");
var slash_1 = __importDefault(require("slash"));
function getAliasSuggestion(importModuleName, aliasConfigs, absoluteDir) {
    var currentAliasConfig = aliasConfigs.find(function (_a) {
        var alias = _a.alias;
        return importModuleName.includes(alias);
    });
    var absoluteModulePath = undefined;
    if (importModuleName.trim().charAt(0) === ".") {
        absoluteModulePath = (0, path_1.join)(absoluteDir, importModuleName);
    }
    else if (currentAliasConfig) {
        absoluteModulePath = importModuleName.replace(currentAliasConfig.alias, currentAliasConfig.path.absolute);
    }
    if (absoluteModulePath) {
        var bestAliasConfig = getBestAliasConfig(aliasConfigs, currentAliasConfig, absoluteModulePath);
        if (bestAliasConfig && bestAliasConfig !== currentAliasConfig) {
            return (0, slash_1["default"])(absoluteModulePath.replace(bestAliasConfig.path.absolute, bestAliasConfig.alias));
        }
    }
}
function getBestAliasConfig(aliasConfigs, currentAlias, absoluteModulePath) {
    return aliasConfigs.reduce(function (currentBest, potentialAlias) {
        var isValidAlias = absoluteModulePath.indexOf(potentialAlias.path.absolute) >= 0;
        var isMoreSpecificAlias = !currentBest ||
            potentialAlias.path.absolute.length >
                currentBest.path.absolute.length;
        return isValidAlias && isMoreSpecificAlias
            ? potentialAlias
            : currentBest;
    }, currentAlias);
}
/**
 * This should match the type definition of ImportAliasOptions
 */
var schemaProperties = {
    aliasConfigPath: {
        description: "Alternative path to look for a TSConfig/JSConfig",
        type: "string"
    },
    aliasImportFunctions: {
        type: "array",
        description: "Function names which supports aliases as its first function." +
            " Examples are `require` and `mock` for `jest.mock`." +
            " Please submit an issue if position parameters needs support",
        items: {
            type: "string"
        },
        "default": ["require", "mock"]
    }
};
var importAliasRule = {
    meta: {
        docs: {
            description: "Suggests shortest matching alias in TSConfig/JSConfig",
            category: "Suggestions",
            recommended: true,
            url: "https://github.com/limegrass/eslint-plugin-import-alias/blob/HEAD/docs/rules/import-alias.md",
            suggestion: true
        },
        fixable: "code",
        schema: [
            {
                type: "object",
                properties: schemaProperties
            },
        ],
        type: "suggestion"
    },
    create: function (context) {
        var cwd = context.getCwd();
        var _a = context.options[0] || {}, aliasConfigPath = _a.aliasConfigPath, _b = _a.aliasImportFunctions, aliasImportFunctions = _b === void 0 ? schemaProperties.aliasImportFunctions["default"] : _b; // No idea what the other array values are
        var aliasesResult = (0, alias_config_1.loadAliasConfigs)(cwd, aliasConfigPath);
        if ("message" in aliasesResult) {
            return {
                Program: function (node) {
                    context.report({ node: node, message: aliasesResult.message });
                }
            };
        }
        var filepath = (0, path_1.resolve)(context.getFilename());
        if (!(0, fs_extra_1.existsSync)(filepath)) {
            return {
                Program: function (node) {
                    context.report({
                        node: node,
                        message: "a filepath must be provided, try with --stdin-filename or call eslint on a file"
                    });
                }
            };
        }
        var absoluteDir = (0, path_1.dirname)(filepath);
        var getReportDescriptor = function (_a, importModuleName) {
            var moduleStart = _a[0], moduleEnd = _a[1];
            // preserve user quote style
            var quotelessRange = [moduleStart + 1, moduleEnd - 1];
            var aliasSuggestion = getAliasSuggestion(importModuleName, aliasesResult, absoluteDir);
            return aliasSuggestion
                ? {
                    message: "import ".concat(importModuleName, " can be written as ").concat(aliasSuggestion),
                    fix: function (fixer) {
                        return fixer.replaceTextRange(quotelessRange, aliasSuggestion);
                    }
                }
                : undefined;
        };
        return {
            CallExpression: function (node) {
                var identifierNode = node.callee.type === "MemberExpression"
                    ? node.callee.property
                    : node.callee;
                if (identifierNode.type === "Identifier" &&
                    aliasImportFunctions.includes(identifierNode.name) &&
                    node.arguments.length) {
                    var importNameNode = node.arguments[0];
                    if (importNameNode.range &&
                        "value" in importNameNode &&
                        typeof importNameNode.value === "string") {
                        var suggestion = getReportDescriptor(importNameNode.range, importNameNode.value);
                        if (suggestion) {
                            context.report(__assign({ node: node }, suggestion));
                        }
                    }
                }
            },
            ImportDeclaration: function (node) {
                if (node.source.range &&
                    typeof node.source.value === "string") {
                    var suggestion = getReportDescriptor(node.source.range, node.source.value);
                    if (suggestion) {
                        context.report(__assign({ node: node }, suggestion));
                    }
                }
            }
        };
    }
};
exports.importAliasRule = importAliasRule;
